#include <opencv2/opencv.hpp>
#include <chrono>
#include <opencv2/core.hpp>
#include "opencv2/imgproc/imgproc.hpp"
#include "opencv2/highgui/highgui.hpp"
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#ifdef __ANDROID__
#include <android/log.h>
#endif

using namespace cv;
using namespace std;

long long int get_now() {
    return chrono::duration_cast<std::chrono::milliseconds>(
            chrono::system_clock::now().time_since_epoch()
    ).count();
}

void platform_log(const char *fmt, ...) {
    va_list args;
    va_start(args, fmt);
#ifdef __ANDROID__
    __android_log_vprint(ANDROID_LOG_VERBOSE, "ndk", fmt, args);
#else
    vprintf(fmt, args);
#endif
    va_end(args);
}

// Avoiding name mangling
extern "C" {
    // Attributes to prevent 'unused' function from being removed and to make it visible
    __attribute__((visibility("default"))) __attribute__((used))
    const char* version() {
        return CV_VERSION;
    }

    Mat getImageCannyBorders( Mat src);
    vector<Point> findBordersPoints(Mat src);
    Mat fourPointsTransform(Mat src, vector<Point>);
    vector<Point> orderPoints(vector<Point> points);
    void swap(Point a, Point b){
        Point c = a;
        a = b;
        b = c;
    }
    bool compareArcLength ( std::vector<cv::Point> contour1, std::vector<cv::Point> contour2 ) {
        double i = fabs( arcLength(cv::Mat(contour1), true));
        double j = fabs( arcLength(cv::Mat(contour2), true));
        return ( i > j );
    }

    /** @function getBorders */
    Mat getImageCannyBorders( Mat src)
    {
        Mat border, edged;
        Mat kernel = getStructuringElement(MORPH_RECT, Size(9.0, 9.0));
        /// Convert it to gray
        cvtColor( src, border, COLOR_BGR2GRAY);
        threshold(border, border, 20.0, 255.0, THRESH_TRIANGLE);
        GaussianBlur( border, border, Size(5,5), 0, 0, BORDER_DEFAULT );
        // Canny edge detector
        Canny(border, edged, 65, 210); 
        dilate(border, border, kernel);

        return border;
    }
    bool compareContourAreas ( std::vector<cv::Point> contour1, std::vector<cv::Point> contour2 ) {
        double i = fabs( contourArea(cv::Mat(contour1)) );
        double j = fabs( contourArea(cv::Mat(contour2)) );
        return ( i > j );
    }
    double angle( cv::Point pt1, cv::Point pt2, cv::Point pt0 ) {
                double dx1 = pt1.x - pt0.x;
                double dy1 = pt1.y - pt0.y;
                double dx2 = pt2.x - pt0.x;
                double dy2 = pt2.y - pt0.y;
                return (dx1*dx2 + dy1*dy2)/sqrt((dx1*dx1 + dy1*dy1)*(dx2*dx2 + dy2*dy2) + 1e-10);
            }

    /** @function findTicket **/
    vector<Point> findBordersPoints(Mat src)
    {
        vector<vector<Point> > contours;	/// Find contours
        vector<Vec4i> hierarchy;
        findContours( src, contours, hierarchy, RETR_TREE, CHAIN_APPROX_SIMPLE);
        
        // Sorted in decreasing order according to its Area
        sort(contours.begin(), contours.end(), compareContourAreas);
        
        int index;
        vector<vector<Point> > contours_poly( contours.size() );
        vector<Point> emptyVector;
        
        //We're only interested in the 5 largest contours
        if(contours.size()>0){  //If contours are detected
        switch (contours.size())
            {
            case 0:
            case 1:
            case 2:
            case 3:
            case 4:
            case 5:
                index = contours.size() -1;
                break;
        
            default:
                index = 4;
                break;
            }
        
       
        for(unsigned int i = 0; i < index; i++ )
        {
            double peri = arcLength(contours[i], true);
            platform_log("Try number: %d.", i);
            // approximate contour with accuracy proportional
            // to the contour perimeter
            approxPolyDP(Mat(contours[i]), contours_poly[i], peri*0.02, true);

            // Note: absolute value of an area is used because
            // area may be positive or negative - in accordance with the
            // contour orientation
            if (contours_poly[i].size() == 4 &&
                fabs(contourArea(Mat(contours_poly[i]))) > 1000 &&
                isContourConvex(Mat(contours_poly[i])))
                {
                    platform_log("succesful try: %d.", i);
                    double maxCosine = 0;

                    for (int j = 2; j < 5; j++)
                        {
                            double cosine = fabs(angle(contours_poly[i][j%4], contours_poly[i][j-2], contours_poly[i][j-1]));
                            maxCosine = MAX(maxCosine, cosine);
                        }

                    if (maxCosine < 0.3)
                        return contours_poly[i];
                }
            
        }}
        return emptyVector;

    }

    vector<Point> orderPoints(vector<Point> points)
    {
        vector<Point> order;
        Point tl(std::numeric_limits<int>::max(), 0);
        Point tr(0,std::numeric_limits<int>::max());
        Point bl(std::numeric_limits<int>::max(),std::numeric_limits<int>::max());
        Point br(std::numeric_limits<int>::min(), std::numeric_limits<int>::min());

        for (int i = 0; i < 4;i++)
        {
            int x = points[i].x;
            int y = points[i].y;
            if (x + y < tl.x + tl.y) tl = points[i];
            if (x - y > tr.x - tr.y) tr = points[i];
            if (x - y < bl.x - bl.y) bl = points[i];
            if (x + y > br.x + br.y) br = points[i];
        }
        order.push_back(tl);
        order.push_back(tr);
        order.push_back(bl);
        order.push_back(br);
        return order;
    }

    Mat fourPointsTransform(Mat src, vector<Point> corners)
    {
        Point2f points[4];
        vector<Point> pv = orderPoints(corners);
        copy(pv.begin(), pv.end(), points);
        Point tl = points[0];
        Point tr = points[1];
        Point bl = points[2];
        Point br = points[3];

        // compute the width of the new image, which will be the
        // maximum distance between bottom-right and bottom-left
        // x-coordiates or the top-right and top-left x-coordinates
        float widthA, widthB, maxWidth;
        float heightA, heightB, maxHeight;

        widthA = sqrt((pow((br.x - bl.x), 2)) + (pow((br.y - bl.y), 2)));
        widthB = sqrt((pow((tr.x - tl.x), 2)) + (pow((tr.y - tl.y), 2)));
        maxWidth = max(int(widthA), int(widthB));

        // compute the height of the new image, which will be the
        // maximum distance between the top-right and bottom-right
        // y-coordinates or the top-left and bottom-left y-coordinates
        heightA = sqrt((pow((tr.x - br.x), 2)) + (pow((tr.y - br.y), 2)));
        heightB = sqrt((pow((tl.x - bl.x), 2)) + (pow((tl.y - bl.y), 2)));
        maxHeight = max(int(heightA), int(heightB));

        // now that we have the dimensions of the new image, construct
        // the set of destination points to obtain a "birds eye view",
        // (i.e. top-down view) of the image, again specifying points
        // in the top-left, top-right, bottom-right, and bottom-left order
        Point2f dts[4];
        dts[0] = Point(0,0);
        dts[1] = Point(maxWidth-1,0);
        dts[2] = Point(0, maxHeight-1);
        dts[3] = Point(maxWidth-1,maxHeight-1);

        Mat warpMatrix = getPerspectiveTransform(points, dts);

        Mat rotated;
        Size size(maxWidth, maxHeight);
        warpPerspective(src, rotated, warpMatrix, size, INTER_LINEAR, BORDER_CONSTANT);
        return rotated;
    }

    /** Get corners detected **/

    __attribute__((visibility("default"))) __attribute__((used))
    const char* pamFindContours(char* path){
        Mat graysrc, src;
        const char* corners; //To transform the vector<Point> to String
        src = imread(path);

        // Get corners
        graysrc = getImageCannyBorders(src);
        vector<Point> borders = findBordersPoints(graysrc);
        if(borders.size() > 0){
            
            borders = orderPoints(borders);
            platform_log("Height: %d", src.rows);
            platform_log("new corners: tl:(%d, %d) tr:(%d, %d) bl:(%d, %d) br:(%d, %d)", borders[0].x, borders[0].y, borders[1].x, borders[1].y, borders[2].x, borders[2].y, borders[3].x, borders[3].y);
            stringstream ss;
            for(size_t i = 0; i < borders.size(); ++i)
            {
                if(i != 0){ss << ";";}
                ss << borders[i].x; ss << ";"; ss << borders[i].y;
            }
            string s = ss.str();
            corners = s.c_str();
            return corners;
        }
        return corners = "noneDetected";
    }


    /** Get image cropped & enhanced **/

    __attribute__((visibility("default"))) __attribute__((used))
    void process_image(char* path, char* outPath, char* points) {


        Mat graysrc, ticketImage, src;
        src = imread(path);

        //Receive cropPoints from string to vector<Point>
        vector<string> vect;
        stringstream ss(points);

        while( ss.good() )
        {
            string substr;
            getline( ss, substr, ',' );
            vect.push_back(substr);
        }

        vector<Point> cropPoints(4); 
        
        cropPoints[0] = Point(stod(vect[0]), stod(vect[1]));
        cropPoints[1] = Point(stod(vect[2]), stod(vect[3]));
        cropPoints[2] = Point(stod(vect[4]), stod(vect[5]));
        cropPoints[3] = Point(stod(vect[6]), stod(vect[7]));

        platform_log("cropPoints: tl:(%d,%d), tr:(%d,%d), bl:(%d,%d), br:(%d,%d)", cropPoints[0].x, cropPoints[0].y,cropPoints[1].x, cropPoints[1].y,
        cropPoints[2].x, cropPoints[2].y,cropPoints[3].x, cropPoints[3].y);

        //FILTERS
        cvtColor(src, graysrc, COLOR_BGR2GRAY); //Obligatory Multi2OneChannel Filter
        graysrc = fourPointsTransform(graysrc, cropPoints); //Crop + Angle transform
        adaptiveThreshold(graysrc,ticketImage, 255,ADAPTIVE_THRESH_MEAN_C,THRESH_BINARY,15,15);
        adaptiveThreshold(ticketImage,ticketImage, 255,ADAPTIVE_THRESH_GAUSSIAN_C,THRESH_BINARY,11,2);
        
        imwrite(outPath, ticketImage);
    
    }
}
